from abc import ABC, abstractmethod
import torch
import torch.nn.functional as Func
import numpy as np 
import math
import logging
import os

import pycvvdp.utils as utils

# I am unsure where it is comming from
# XYZ_to_LMS2006 = (
#   ( 0.185081950317403,   0.584085520715683,  -0.024070430377029 ),
#   (-0.134437959455888,   0.405757045863316,   0.035826598616685 ),
#   ( 0.000790172132780,  -0.000913083981083,   0.019850963261241 ) )

XYZ_to_LMS2006 = (
   ( 0.187596268556126,   0.585168649077728,  -0.026384263306304 ),
   (-0.133397430663221,   0.405505777260049,   0.034502127690364 ),
   (0.000244379021663,  -0.000542995890619,   0.019406849066323 ) )

LMS2006_to_DKLd65 = (
  (1.000000000000000,   1.000000000000000,                   0),
  (1.000000000000000,  -2.311130179947035,                   0),
  (-1.000000000000000,  -1.000000000000000,  50.977571328718781) )

XYZ_to_RGB2020 = (  (1.716502508360628, -0.355584689096764,  -0.253375213570850), \
                    (-0.666625609145029,   1.616446566522207,   0.015775479726511), \
                    (0.017655211703087,  -0.042810696059636,   0.942089263920533) )

XYZ_to_RGB709 = (   ( 3.2406, -1.5372, -0.4986), \
                    (-0.9689,  1.8758,  0.0415), \
                    (0.0557, -0.2040,  1.0570) )

def lms2006_to_dkld65( img ):
    M = torch.as_tensor( LMS2006_to_DKLd65, dtype=img.dtype, device=img.device)

    ABC = torch.empty_like(img)  # ABC represents any linear color space
    # To avoid permute (slow), perform separate dot products
    for cc in range(3):
        ABC[...,cc,:,:,:] = torch.sum(img*(M[cc,:].view(1,3,1,1,1)), dim=-4, keepdim=True)
    return ABC

def lin2pq( L ):
    """ Convert from absolute linear values (between 0.005 and 10000) to PQ-encoded values V (between 0 and 1)
    """
    Lmax = 10000
    #Lmin = 0.005
    n    = 0.15930175781250000
    m    = 78.843750000000000
    c1   = 0.83593750000000000
    c2   = 18.851562500000000
    c3   = 18.687500000000000
    im_t = (torch.clip(L,0,Lmax)/Lmax) ** n
    V  = ((c2*im_t + c1) / (1+c3*im_t)) ** m
    return V

def pq2lin( V ):
    """ Convert from PQ-encoded values V (between 0 and 1) to absolute linear values (between 0.005 and 10000)
    """
    Lmax = 10000
    n    = 0.15930175781250000
    m    = 78.843750000000000
    c1   = 0.83593750000000000
    c2   = 18.851562500000000
    c3   = 18.687500000000000

    im_t = torch.pow(V,1/m)
    L = Lmax * torch.pow((im_t-c1).clamp(min=0)/(c2-c3*im_t), 1/n)
    return L

# Convert pixel values to linear RGB using sRGB non-linearity
# 
# L = srgb2lin( p )
#
# p - pixel values (between 0 and 1)
# L - relative linear RGB (or luminance), normalized to the range 0-1
def srgb2lin( p ):
    L = torch.where(p > 0.04045, ((p + 0.055) / 1.055)**2.4, p/12.92)
    return L


# Convert pixel values to linear using the Rec. 2100 HLG non-linearity
#
# rgb_d = hlg2lin( rgb, gamma )
#
# rgb   - pixel values (between 0 and 1)
# rgb_d - relative linear RGB (or luminance), normalized to the range 0-1
def hlg2lin( rgb, gamma ):
    # using formula from table 5 of
    # https://www.itu.int/dms_pubrec/itu-r/rec/bt/R-REC-BT.2100-1-201706-S!!PDF-E.pdf

    a = 0.17883277
    b = 1 - 4 * a
    c = 0.5 - a * math.log(4 * a)

    # inverse OETF
    rgb_s = torch.where(
        rgb <= 0.5,
        torch.pow(rgb, 2) / 3.0,
        (torch.exp((rgb - c) / a) + b) / 12.0
    )

    # apply OOTF
    Y_s = 0.2627 * rgb_s[:, 0] + 0.6780 * rgb_s[:, 1] + 0.0593 * rgb_s[:, 2]
    rgb_d = (Y_s ** (gamma - 1)).unsqueeze(1) * rgb_s

    return rgb_d

class vvdp_display_photometry:

    def __init__( self, source_colorspace='sRGB', config_paths=[] ):

        colorspaces_file = utils.config_files.find( "color_spaces.json", config_paths )
        colorspaces = utils.json2dict(colorspaces_file)

        if not source_colorspace in colorspaces:
            raise RuntimeError( f'Color space: "{source_colorspace}" not found in "{colorspaces_file}"' )

        if 'RGB2X' in colorspaces[source_colorspace]: # luminance will not have colour space primaries
            self.rgb2xyz_list = [colorspaces[source_colorspace]['RGB2X'], colorspaces[source_colorspace]['RGB2Y'], colorspaces[source_colorspace]['RGB2Z'] ]
        self.EOTF = colorspaces[source_colorspace]['EOTF']


    # Transforms gamma-encoded pixel values V, which must be in the range
    # 0-into absolute linear colorimetric values emitted from
    # the display.
    @abstractmethod
    def forward( self, V ):
        pass

    # Print the display specification    
    @abstractmethod
    def print( self ):
        pass

    # @classmethod
    # def default_model_file( cls ):
    #     return os.path.join(os.path.dirname(__file__), "fvvdp_data/display_models.json")

    @classmethod
    def list_displays( cls, config_paths ):
        models_file = utils.config_files.find( "display_models.json", config_paths )

        logging.info( f"JSON file with display models: {models_file}" )

        models = utils.json2dict(models_file)

        for display_name in models:
            dm = vvdp_display_photometry.load(display_name, config_paths)
            dm.print()

    @classmethod
    def load( cls, display_name, config_paths ):
        models_file = utils.config_files.find( "display_models.json", config_paths )

        models = utils.json2dict(models_file)

        if not display_name in models:
            logging.error(f"Display model: '{display_name}' not found in '{models_file}'")
            raise RuntimeError( 'Display model not found' )

        model = models[display_name]

        Y_peak = model["max_luminance"]

        if "colorspace" in model:
            colorspace = model["colorspace"]
        else:
            colorspace = 'sRGB'

        if "min_luminance" in model:
            contrast = Y_peak/model["min_luminance"]
        elif "contrast" in model:
            contrast = model["contrast"]
        else:
            contrast = 500

        # Ambient light
        if "E_ambient" in model:
            E_ambient = model["E_ambient"]
        else:
            E_ambient = 0

        # Reflectivity of the display panel
        if "k_refl" in model: 
            k_refl = model["k_refl"]
        else:
            k_refl = 0.005

        # Exposure
        if "exposure" in model:
            exposure = model["exposure"]
        else:
            exposure = 1

        obj = vvdp_display_photo_eotf( Y_peak, contrast=contrast, source_colorspace=colorspace, E_ambient=E_ambient, k_refl=k_refl, name=display_name, exposure=exposure, config_paths=config_paths)
        obj.full_name = model["name"]
        obj.short_name = display_name

        return obj

    # Transform content from its source color space (typically display-encoded RGB) into 
    # the colorimetric values of light emmitted from the display and then into the target color
    # space used by a metric.
    def source_2_target_colorspace(self, I_src, target_colorspace):        

        if target_colorspace in ['display_encoded_01', 'display_encoded_dmax', 'display_encoded_100nit']: # if a display-encoded frame is requested

            # Special case - if PQ, we still want to use PU21, as it should be marginally better
            if self.is_input_display_encoded() and not (isinstance( self, vvdp_display_photo_eotf) and self.EOTF == 'PQ'):
                I_target = I_src # no need to do anything
            else:
                # Otherwise, we need to PU-encode the frame
                if not hasattr( self, "PU" ):
                    self.PU = utils.PU()

                if target_colorspace == 'display_encoded_01':
                    PU_max = self.PU.encode(torch.as_tensor(10000.0))
                elif target_colorspace == 'display_encoded_100nit':
                    PU_max = self.PU.encode(torch.as_tensor(100.0)) # White diffuse of 100 nit will be mapped to 1
                else:
                    PU_max = self.PU.encode(torch.as_tensor(self.get_peak_luminance()))
                
                I_lin = self.forward( I_src )
                I_target = self.PU.encode(I_lin) / PU_max 
        else:
            # Apply forward display model to get absolute linear values
            I_lin = self.forward( I_src )

            is_color = (I_src.shape[-4]==3)
            if is_color:
                I_target = self.linear_2_target_colorspace(I_lin, target_colorspace)
            else:
                I_target = I_lin

        return I_target

    # Transform frame/image from native linear color space to the target color space.
    # Internal, do not use. 
    def linear_2_target_colorspace(self, RGB_lin, target_colorspace):        
        if hasattr(self, "rgb2xyz"):
            rgb2xyz = self.rgb2xyz
        else:
            rgb2xyz = torch.tensor( self.rgb2xyz_list, dtype=RGB_lin.dtype, device=RGB_lin.device )
            self.rgb2xyz = rgb2xyz

        if target_colorspace=="Y":
            return torch.sum(RGB_lin*(rgb2xyz[1,:].view(1,3,1,1,1)), dim=-4, keepdim=True)
        else:
            if target_colorspace=="XYZ":
                rgb2abc = rgb2xyz
            elif target_colorspace=="LMS2006":
                rgb2abc = torch.as_tensor( XYZ_to_LMS2006, dtype=RGB_lin.dtype, device=RGB_lin.device) @ rgb2xyz
            elif target_colorspace=="DKLd65":
                rgb2abc = torch.as_tensor( LMS2006_to_DKLd65, dtype=RGB_lin.dtype, device=RGB_lin.device) @ torch.as_tensor( XYZ_to_LMS2006, dtype=RGB_lin.dtype, device=RGB_lin.device) @ rgb2xyz
            elif target_colorspace=="RGB709":
                rgb2abc = torch.as_tensor( XYZ_to_RGB709, dtype=RGB_lin.dtype, device=RGB_lin.device) @ rgb2xyz
            elif target_colorspace=="RGB2020" or target_colorspace=="RGB2020pq":
                rgb2abc = torch.as_tensor( XYZ_to_RGB2020, dtype=RGB_lin.dtype, device=RGB_lin.device) @ rgb2xyz
            elif target_colorspace=="logLMS_DKLd65":
                rgb2abc = torch.as_tensor( XYZ_to_LMS2006, dtype=RGB_lin.dtype, device=RGB_lin.device) @ rgb2xyz
            else:
                raise RuntimeError( f"Unknown colorspace '{target_colorspace}'" )

            ABC = torch.empty_like(RGB_lin)  # ABC represents any linear color space
            # To avoid permute (slow), perform separate dot products
            for cc in range(3):
                ABC[:,cc:(cc+1),:,:,:] = torch.sum(RGB_lin*(rgb2abc[cc,:].view(1,3,1,1,1)), dim=-4, keepdim=True)

            if target_colorspace=="logLMS_DKLd65":
                ABC = lms2006_to_dkld65( torch.log10(ABC) )
            elif target_colorspace=="RGB2020pq":
                ABC = lin2pq(ABC)

            return ABC

class vvdp_display_photo_eotf(vvdp_display_photometry): 
    # Display model with several EOTF, to simulate both SDR and HDR displays
    #
    # dm = vvdp_display_photo_eotf( Y_peak, contrast, EOTF, gamma, E_ambient, k_refl )
    #
    # Parameters (default value shown in []):
    # Y_peak - display peak luminance in cd/m^2 (nit), e.g. 200 for a typical
    #          office monitor, 1000 for an HDR display, ...
    # contrast - [1000] the contrast of the display. The value 1000 means
    #          1000:1
    # source_colorspace - color space from colorspaces.json. colorspace entry includes EOTF, 
    #          but it can be overriden using EOTF parameter.
    # EOTF - 'sRGB', 'PQ', 'linear' or a string with a numeric value, such as "2.2", for gamma 2.2. 
    #        This parameter will overwrite the EOTF attribute in the JSON file with corresponding 'source_colorspace'.
    # E_ambient - [0] ambient light illuminance in lux, e.g. 600 for bright
    #         office
    # k_refl - [0.005] reflectivity of the display screen
    # exposure - [1] exposure of the content. The colour in the linear colour space is multipled by this constant.
    #
    # For more details on the GOG display model, see:
    # https://www.cl.cam.ac.uk/~rkm38/pdfs/mantiuk2016perceptual_display.pdf
    #
    # Copyright (c) 2010-2022, Rafal Mantiuk
    def __init__( self, Y_peak, contrast = 1000, source_colorspace='sRGB', EOTF=None, E_ambient = 0, k_refl = 0.005, exposure=1, name=None, config_paths=[] ):
            
        super().__init__(source_colorspace=source_colorspace, config_paths=config_paths)
        if not EOTF is None: 
            self.EOTF = EOTF

        self.Y_peak = Y_peak            
        self.contrast = contrast
        self.E_ambient = E_ambient
        self.k_refl = k_refl
        self.name = name    
        self.exposure = exposure

    # Say whether the input frame is display-encoded. False if it is linear. 
    def is_input_display_encoded(self):
        # Is not display encoded if EOTF is "linear"
        return (self.EOTF!='linear')

    def __eq__(self, other): 
        if not isinstance(other, self.__class__):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.Y_peak == other.Y_peak \
            and self.contrast == other.contrast \
            and self.EOTF == other.EOTF \
            and self.E_ambient == other.E_ambient \
            and self.k_refl == other.k_refl \
            and self.exposure == other.exposure

    # Transforms display-encoded pixel values V, which must be in the range
    # 0-1 into absolute linear colorimetric values emitted from
    # the display.
    def forward( self, V ):
        
        if self.EOTF != 'linear' and ((V>1).flatten().any() or (V<0).flatten().any()):
            logging.warning("Pixel outside the valid range 0-1")
            V = V.clamp( 0., 1. )
            
        Y_black, Y_refl = self.get_black_level()
                
        if self.EOTF=='sRGB':
            if self.exposure == 1:
                L = (self.Y_peak-Y_black)*srgb2lin(V) + Y_black + Y_refl
            else:
                L = (self.Y_peak-Y_black)*(srgb2lin(V)*self.exposure).clip(0., 1.) + Y_black + Y_refl
        elif self.EOTF=='PQ':
            L = (pq2lin( V )*self.exposure).clip(0.005, self.Y_peak) + Y_black + Y_refl #TODO: soft clipping
        elif self.EOTF=='linear':
            L = (V*self.exposure).clip(max(0.005, Y_black), self.Y_peak) + Y_refl #TODO: soft clipping
        elif self.EOTF=='HLG':
            gamma = 1.2
            if self.Y_peak > 1000:
                # The correction term "- 0.07623 * math.log10(self.E_ambient / 5)" comes from BBC Research & Development White Paper WHP 369
                # https://downloads.bbc.co.uk/rd/pubs/whp/whp-pdf-files/WHP369.pdf
                gamma = 1.2 + 0.42 * math.log10(self.Y_peak / 1000) - 0.07623 * math.log10(self.E_ambient / 5)
            if self.exposure == 1:
                L = (self.Y_peak-Y_black)*hlg2lin(V, gamma) + Y_black + Y_refl
            else:
                L = (self.Y_peak-Y_black)*(hlg2lin(V, gamma)*self.exposure).clip(0., 1.) + Y_black + Y_refl
        elif self.EOTF[0].isnumeric(): # if the first char is numeric -> gamma
            gamma = float(self.EOTF)
            L = (self.Y_peak-Y_black)*(torch.pow(V, gamma)*self.exposure).clip(0., 1.) + Y_black + Y_refl
        else:
            raise RuntimeError( f"Unknown EOTF '{self.EOTF}'" )        
        return L
        

    def get_peak_luminance( self ):
        return self.Y_peak

    # Get the black level and the light reflected from the display
    def get_black_level( self ):
        Y_refl = self.E_ambient/math.pi*self.k_refl  # Reflected ambient light            
        Y_black = self.Y_peak/self.contrast

        return Y_black, Y_refl

    # Print the display specification    
    def print( self ):
        Y_black, Y_refl = self.get_black_level()
        
        logging.info( 'Photometric display model: {}'.format(self.name) )
        logging.info( '  Peak luminance: {} cd/m^2'.format(self.Y_peak) )
        logging.info( '  EOTF: {}'.format(self.EOTF) )
        logging.info( '  Contrast - theoretical: {}:1'.format( round(self.contrast) ) )
        logging.info( '  Contrast - effective: {}:1'.format( round(self.Y_peak/(Y_black+Y_refl)) ) )
        logging.info( '  Ambient light: {} lux'.format( self.E_ambient ) )
        logging.info( '  Display reflectivity: {}%'.format( self.k_refl*100 ) )
    


# Use this class to compute the effective resolution of a display in pixels
# per degree (ppd). The class accounts for the change in the projection
# when looking at large FOV displays (e.g. VR headsets) at certain
# eccentricity.
#
# The class is also useful for computing the size of a display in meters
# and visual degrees. Check 'display_size_m' and 'display_size_deg' class
# properties for that.
#
# R = fvvdp_display_geometry(resolution, distance_m=None, distance_display_heights=None, 
#                            fov_horizontal=None, fov_vertical=None, fov_diagonal=None, 
#                            diagonal_size_inches=None)
#
# resolution is the 2-element touple with the pixel resolution of the
# display: (horizontal_resolutution, vertical_resolutution)
# distance_m - viewing distance in meters
# distance_display_heights - viewing distance in the heights of a display
# fov_horizontal - horizontal field of view of the display in degrees
# fov_vertical - vertical field of view of the display in degrees
# fov_diagonal - diagonal field of view of the display in degrees
# diagonal_size_inches - display diagonal resolution in inches
#
# Examples:
# # HTC Pro
# # Note that the viewing distance must be specified even though the resolution
# # and 'fov_diagonal' are enough to find pix_per_deg.
# R = fvvdp_display_geometry( (1440, 1600), distance_m=3, fov_diagonal=110 )
# R.get_ppd( torch.tensor( [0, 10, 20, 30, 40])) # pix per deg at the given eccentricities
#
# # 30" 4K monitor seen from 0.6 meters
# R = fvvdp_display_geometry( (3840, 2160), diagonal_size_inches=30, distance_m=0.6 )
# R.get_ppd()
#
# # 47" SIM2 display seen from 3 display heights
# R = fvvdp_display_geometry( (1920, 1080), diagonal_size_inches=47, distance_display_heights=3 )
# R.get_ppd()
#
# Some information about the effective FOV of VR headsets
# http://www.sitesinvr.com/viewer/htcvive/index.html
class vvdp_display_geometry:

    def __init__(self, resolution, distance_m=None, distance_display_heights=None, fov_horizontal=None, fov_vertical=None, fov_diagonal=None, diagonal_size_inches=None, ppd=None) -> None:

        self.resolution = resolution
        
        ar = resolution[0]/resolution[1] # width/height
        
        if not ppd is None:
            self.fixed_ppd = ppd
            return
        else:
            self.fixed_ppd = None

        if not diagonal_size_inches is None:
            height_mm = math.sqrt( (diagonal_size_inches*25.4)**2 / (1+ar**2) )
            self.display_size_m = (ar*height_mm/1000, height_mm/1000)
                
        if (not distance_m is None) and (not distance_display_heights is None):
            raise RuntimeError( 'You can pass only one of: ''distance_m'', ''distance_display_heights''.' )
        
        if not distance_m is None:
            self.distance_m = distance_m;
        elif not distance_display_heights is None:
            if not hasattr( self, "display_size_m" ):
                raise RuntimeError( 'You need to specify display diagonal size ''diagonal_size_inches'' to specify viewing distance as ''distance_display_heights'' ' )
            self.distance_m = distance_display_heights * self.display_size_m[1]
        elif (not fov_horizontal is None) or (not fov_vertical is None) or (not fov_diagonal is None):
            # Default viewing distance for VR headsets
            self.distance_m = 3
        else:
            raise RuntimeError( 'Viewing distance must be specified as ''distance_m'' or ''distance_display_heights''.' )
        
        if ((not fov_horizontal is None) + (not fov_vertical is None) + (not fov_diagonal is None)) > 1:
            raise RuntimeError( 'You can pass only one of ''fov_horizontal'', ''fov_vertical'', ''fov_diagonal''. The other dimensions are inferred from the resolution assuming that the pixels are square.' )
        
        if not fov_horizontal is None:
            width_m = 2*math.tan( math.radians(fov_horizontal/2) )*self.distance_m
            self.display_size_m = (width_m, width_m/ar)
        elif not fov_vertical is None:
            height_m = 2*math.tan( math.radians(fov_vertical/2) )*self.distance_m
            self.display_size_m = (height_m*ar, height_m)
        elif not fov_diagonal is None:
            # Note that we cannot use Pythagorean theorem on degs -
            # we must operate on a distance measure
            # This is incorrect: height_deg = p.Results.fov_diagonal / sqrt( 1+ar^2 );
            
            distance_px = math.sqrt(self.resolution[0]**2 + self.resolution[1]**2) / (2.0 * math.tan( math.radians(fov_diagonal*0.5)) )
            height_deg = math.degrees(math.atan( self.resolution[1]/2 / distance_px ))*2
            
            height_m = 2*math.tan( math.radians(height_deg/2) )*self.distance_m
            self.display_size_m = (height_m*ar, height_m)
        
        self.display_size_deg = ( 2 * math.degrees(math.atan( self.display_size_m[0] / (2*self.distance_m) )), \
                                  2 * math.degrees(math.atan( self.display_size_m[1] / (2*self.distance_m) )) )

    def __eq__(self, other): 
        if not isinstance(other, self.__class__):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.resolution == other.resolution \
            and self.distance_m == other.distance_m \
            and self.display_size_m == other.display_size_m

    # Get the number of pixels per degree
    #
    # ppd = R.get_ppd()
    # ppd = R.get_ppd(eccentricity)
    #
    # eccentricity is the viewing angle from the center in degrees. If
    # not specified, the central ppd value (for 0 eccentricity) is
    # returned.
    def get_ppd(self, eccentricity = None):
        
        # if ~isempty( dr.fixed_ppd )
        #     ppd = dr.fixed_ppd;
        #     return;
        # end
        
        if not self.fixed_ppd is None:
            return self.fixed_ppd

        # pixel size in the centre of the display
        pix_deg = 2*math.degrees(math.atan( 0.5*self.display_size_m[0]/self.resolution[0]/self.distance_m ))
        
        base_ppd = 1/pix_deg
        
        if eccentricity is None:
            return base_ppd
        else:
            delta = pix_deg/2
            tan_delta = math.tan(math.radians(delta))
            tan_a = torch.tan( torch.deg2rad(eccentricity) )
            
            ppd = base_ppd * (torch.tan(torch.deg2rad(eccentricity+delta))-tan_a)/tan_delta
            return ppd


    # Convert pixel positions into eccentricities for the given
    # display
    #
    # resolution_pix - image resolution as [width height] in pix
    # x_pix, y_pix - pixel coordinates generated with meshgrid,
    #   pixels indexed from 0
    # gaze_pix - [x y] of the gaze position, in pixels
    def pix2eccentricity( self, resolution_pix, x_pix, y_pix, gaze_pix ):
                        
        if not self.fixed_ppd is None:
            ecc = torch.sqrt( (x_pix-gaze_pix[0])**2 + (y_pix-gaze_pix[1])**2 )/self.fixed_ppd
        else:
            # Position the image in the centre
            shift_to_centre = -resolution_pix/2
            x_pix_rel = x_pix+shift_to_centre[0]
            y_pix_rel = y_pix+shift_to_centre[1]
            
            x_m = x_pix_rel * self.display_size_m[0] / self.resolution[0]
            y_m = y_pix_rel * self.display_size_m[1] / self.resolution[1]
            
            device = x_pix.device

            gaze_m = (gaze_pix + shift_to_centre) * torch.tensor(self.display_size_m) / torch.tensor(self.resolution)
            gaze_deg = torch.rad2deg(torch.atan( gaze_m/self.distance_m ))
            
            ecc = torch.sqrt( (torch.rad2deg(torch.atan(x_m/self.distance_m))-gaze_deg[0])**2 + (torch.rad2deg(torch.atan(y_m/self.distance_m))-gaze_deg[1])**2 )
        
        return ecc
        
    def get_resolution_magnification( self, eccentricity ):
            # Get the relative magnification of the resolution due to
            # eccentricity.
            # 
            # M = R.get_resolution_magnification(eccentricity)
            # 
            # eccentricity is the viewing angle from the center to the fixation point in degrees.
            
            if not self.fixed_ppd is None:
                M = torch( (1), device=eccentricity.device )
            else:            
                eccentricity = torch.minimum( eccentricity, torch.tensor((89.9)) ) # To avoid singulatities
                
                # pixel size in the centre of the display
                pix_rad = 2*math.atan( 0.5*self.display_size_m[0]/self.resolution[0]/self.distance_m )
                
                delta = pix_rad/2
                tan_delta = math.tan(delta)
                tan_a = torch.tan( torch.deg2rad(eccentricity) )
                
                M = (torch.tan(torch.deg2rad(eccentricity)+delta)-tan_a)/tan_delta

            return M

    def print(self):
        logging.info( 'Geometric display model:' )
        if hasattr( self, "fixed_ppd" ):
            logging.info( '  Fixed pixels-per-degree: {}'.format(self.fixed_ppd) )
        else:
            logging.info( '  Resolution: {w} x {h} pixels'.format( w=self.resolution[0], h=self.resolution[1] ) )
            logging.info( '  Display size: {w:.1f} x {h:.1f} cm'.format( w=self.display_size_m[0]*100, h=self.display_size_m[1]*100) )
            logging.info( '  Display size: {w:.2f} x {h:.2f} deg'.format( w=self.display_size_deg[0], h=self.display_size_deg[1] ) )
            logging.info( '  Viewing distance: {d:.3f} m'.format(d=self.distance_m) )
            logging.info( '  Pixels-per-degree (center): {ppd:.2f}'.format(ppd=self.get_ppd()) )

    @classmethod
    def load( cls, display_name, config_paths=[] ):

        models_file = utils.config_files.find( "display_models.json", config_paths )
        models = utils.json2dict(models_file)

        if not display_name in models:
            logging.error(f"Display model: '{display_name}' not found in '{models_file}'")
            raise RuntimeError( 'Display model not found' )

        model = models[display_name]

        assert "resolution" in model

        inches_to_meters = 0.0254

        W, H = model["resolution"]

        if "pixels_per_degree" in model:
            obj = vvdp_display_geometry( (W, H), ppd=model["pixels_per_degree"])
        else:
            if "fov_diagonal" in model: fov_diagonal = model["fov_diagonal"]
            else:                       fov_diagonal = None

            if   "viewing_distance_meters" in model: distance_m = model["viewing_distance_meters"]
            elif "viewing_distance_inches" in model: distance_m = model["viewing_distance_inches"] * inches_to_meters
            else:                                    distance_m = None

            if   "diagonal_size_meters" in model: diag_size_inch = model["diagonal_size_meters"] / inches_to_meters
            elif "diagonal_size_inches" in model: diag_size_inch = model["diagonal_size_inches"] 
            else:                                 diag_size_inch = None

            obj = vvdp_display_geometry( (W, H), distance_m=distance_m, fov_diagonal=fov_diagonal, diagonal_size_inches=diag_size_inch)
        return obj




import torch

from pycvvdp.video_source import *
from pycvvdp.vq_metric import *
from pycvvdp.video_writer import VideoWriter

try:
    # This may fail if OpenEXR is not installed. To install,
    # ubuntu: sudo apt install libopenexr-dev
    # mac: brew install openexr
    import pyexr
    pyexr_imported = True
except ImportError as e:
    # Imageio's imread is unreliable for OpenEXR images
    # See https://github.com/imageio/imageio/issues/517
    pyexr_imported = False


def tensor_to_numpy_image(T):
    return torch.squeeze(T.permute((3,4,1,0,2)), dim=(3,4)).cpu().numpy()

"""
A fake metric that writes the output of a display model to either HDR video or OpenEXR images/frames. This is useful for checking and debugging display models. 
"""
class dm_preview(vq_metric):

    def __init__(self, output_exr=False, side_by_side=False, display_name="standard_4k", display_photometry=None, device=None, verbose=False):
        # Use GPU if available
        if device is None:
            if torch.cuda.is_available() and torch.cuda.device_count()>0:
                self.device = torch.device('cuda:0')
            else:
                self.device = torch.device('cpu')
        else:
            self.device = device

        self.output_exr = output_exr
        self.side_by_side = side_by_side
        self.set_display_model( display_name=display_name, display_photometry=display_photometry )        
        self.verbose = verbose


    def predict_video_source(self, vid_source, frame_padding="replicate"):

        _, _, N_frames = vid_source.get_video_size()
        
        batch_sz = vid_source.get_batch_size()
        assert batch_sz>1, 'DM-preview does not work with batches'

        write_exr = self.output_exr or N_frames==1

        if write_exr:
            colorspace = 'RGB709'
        else:
            colorspace = 'RGB2020pq'
            fps = vid_source.get_frames_per_second()
            test_vw = VideoWriter(self.base_fname + "-test.mp4", hdr_mode=True, fps=fps, codec='h265', verbose=self.verbose)
            if not self.side_by_side:
                ref_vw = VideoWriter(self.base_fname + "-reference.mp4", hdr_mode=True, fps=fps, codec='h265', verbose=self.verbose)

        for ff in range(N_frames):
            T = vid_source.get_test_frame(ff, device=self.device, colorspace=colorspace)
            R = vid_source.get_reference_frame(ff, device=self.device, colorspace=colorspace)

            if self.side_by_side:
                concat_dim = -1 if T.shape[-1]<T.shape[-2] else -2
                T = torch.concatenate( (T, R), dim=concat_dim )

            frame_no = f"-{ff:04d}" if N_frames>1 else ""

            if write_exr:
                pyexr.write(self.base_fname + frame_no + "-test.exr", tensor_to_numpy_image(T))
                if not self.side_by_side:
                    pyexr.write(self.base_fname + frame_no + "-reference.exr", tensor_to_numpy_image(R))
            else:
                test_vw.write_frame_rgb(tensor_to_numpy_image(T))
                if not self.side_by_side:
                    ref_vw.write_frame_rgb(tensor_to_numpy_image(R))
        
        if not write_exr:
            test_vw.close()
            if not self.side_by_side:
                ref_vw.close()

        return torch.as_tensor(-1, device=self.device), None

    def short_name(self):
        return "dm-preview"

    def quality_unit(self):
        return ""

class dm_preview_exr(dm_preview):
    def __init__(self, **kwargs):
        super().__init__(output_exr=True, **kwargs)

class dm_preview_sbs(dm_preview):
    def __init__(self, **kwargs):
        super().__init__(side_by_side=True, **kwargs)

class dm_preview_exr_sbs(dm_preview):
    def __init__(self, **kwargs):
        super().__init__(side_by_side=True, output_exr=True, **kwargs)

register_metric( dm_preview )
register_metric( dm_preview_sbs )
register_metric( dm_preview_exr )
register_metric( dm_preview_exr_sbs )




import torch
import pycvvdp.utils as utils

from interp import interp1q, batch_interp1d

class castleCSF:

    def __init__(self, csf_version, device, config_paths=[]):
        self.device = device
        csf_lut_file = utils.config_files.find( f"csf_lut_{csf_version}.json", config_paths )
        csf_lut = utils.json2dict(csf_lut_file)

        self.log_L_bkg = torch.log10( torch.as_tensor(csf_lut["L_bkg"], device=device) )
        self.log_rho = torch.log10( torch.as_tensor(csf_lut["rho"], device=device) )
        self.omega = csf_lut["omega"]

        self.logS = []
        for oo in range(2): # For each temp frequency
            self.logS.append([])
            ch_num = 3 if oo==0 else 1
            for cc in range(ch_num):
                field_name = f"o{self.omega[oo]}_c{cc+1}"
                self.logS[oo].append( torch.as_tensor(csf_lut[field_name], device=device) )

        self.logS_rho = {}


    def sensitivity(self, rho, omega, logL_bkg, cc, sigma):
        # rho - spatial frequency
        # omega - temporal frequency
        # L_bkg - background luminance
        # sigma - radius of spatial integration (Gaussian envelope)

        # Which LUT to use
        oo = 0 if omega==0 else 1
        logS = self.logS[oo][cc]

        # First interpolate between spatial frequencies rho
        rho_str = f"o{oo}_c{cc}_rho{rho}"
        if rho_str in self.logS_rho: # Check if it is cached
            logS_r = self.logS_rho[rho_str]
        else:
            N = self.log_L_bkg.numel()
            logS_r = torch.empty((N), device=self.device)
            logS_r = batch_interp1d(torch.log10(torch.as_tensor(rho, device=self.device, dtype=torch.float32)).expand(N), self.log_rho, logS)
            self.logS_rho[rho_str] = logS_r        

        # Then, interpolate across luminance levels    
        S = 10**interp1q( self.log_L_bkg, logS_r, logL_bkg )

        return S

    def update_device( self, device ):
        self.device = device
        self.log_L_bkg = self.log_L_bkg.to(device)
        self.log_rho = self.log_rho.to(device)

        for oo in range(2): # For each temp frequency
            ch_num = 3 if oo==0 else 1
            for cc in range(ch_num):
                self.logS[oo][cc] = self.logS[oo][cc].to(device)



# Dump intermediate channel data for debugging and visualization
import torch
import math
import os
import logging

# For debugging only
#from gfxdisp.pfs.pfs_torch import pfs_torch

from pycvvdp.video_writer import VideoWriter
from pycvvdp.lpyr_dec import lpyr_dec_2

DKLd65_to_RGB = ( (0.926502308187832, 0.960842501786725, 0.940315924461593),
   (6.448879567147620, -2.074854167137361, 0.100486265553559),
   (0.181670434983238, -0.190064026530768, 1.080345193424545) )

def dkld65_to_rgb( img ):
    M = torch.as_tensor( DKLd65_to_RGB, dtype=img.dtype, device=img.device).transpose(0,1)

    ABC = torch.empty_like(img)  # ABC represents any linear colour space
    # To avoid permute (slow), perform separate dot products
    for cc in range(3):
        ABC[...,cc,:,:,:] = torch.sum(img*(M[cc,:].view(1,3,1,1,1)), dim=-4, keepdim=True)
    return ABC

# Round an integer up to be divisible by 8
def ceil8(x):
    return int(math.ceil(x/8))*8

class DumpChannels:
    def __init__(self, dump_temp_ch=True, dump_lpyr=True, dump_diff=True, output_dir=None):
        self.vw_channels = None
        self.do_dump_temp_ch = dump_temp_ch
        self.do_dump_lpyr = dump_lpyr
        self.do_dump_diff = dump_diff
        self.output_dir = output_dir if output_dir else "."

    def open(self, fps):

        assert fps>0, "This feature currently works only on video"

        if self.do_dump_temp_ch:
            fname = os.path.join( self.output_dir, "temp_channels.mp4" )
            logging.info( f"Writing temporal channels to '{fname}'" )
            self.vw_channels = VideoWriter( fname, fps=fps, verbose=False )
        else:
            self.vw_channels = None

        self.max_V = None

        if self.do_dump_lpyr:
            fname = os.path.join( self.output_dir, "lpyr.mp4" )
            logging.info( f"Writing Laplacian pyramids to '{fname}'" )
            self.vw_lpyr = VideoWriter( fname, fps=fps )
        else:
            self.vw_lpyr = None

        if self.do_dump_diff:
            fname = os.path.join( self.output_dir, "diff.mp4" )
            logging.info( f"Writing visual differences to '{fname}'" )
            self.vw_diff = VideoWriter( fname, fps=fps )
            self.diff_pyr = None
        else:
            self.vw_diff = None


    def dump_temp_ch( self, R ):
        # Order: test-sustained-Y, ref-sustained-Y, test-rg, ref-rg, test-yv, ref-yv, test-transient-Y, ref-transient-Y
        # Images do not have the two last channels
        # R = torch.zeros((1, 8, cur_block_N_frames, height, width), device=self.device)

        if not self.do_dump_temp_ch:
            return

        # We dump only test at this moment
        white_dkl = torch.as_tensor( [1, 0.003775328226986, 0.010327227989383], device=R.device )
        ach_sust = R[0:1,0:1,...]
        ach_sust_rgb = dkld65_to_rgb( torch.cat( [ ach_sust, white_dkl[1].expand_as(ach_sust), white_dkl[2].expand_as(ach_sust) ], dim=1 ) )

        if not self.max_V:
            self.max_V = ach_sust_rgb.max()

        gray = white_dkl.view([1,3,1,1,1]) * (self.max_V/4)
        ach_trans = R[0:1,6:7,...]
        ach_trans_rgb = dkld65_to_rgb( torch.cat( [ ach_trans, white_dkl[1].expand_as(ach_sust), white_dkl[2].expand_as(ach_sust) ], dim=1 )+gray )
        rg = R[0:1,2:3,...]
        rg_rgb = dkld65_to_rgb( torch.cat( [ white_dkl[0].expand_as(rg), rg, white_dkl[2].expand_as(rg) ], dim=1 )+gray )
        yv = R[0:1,4:5,...]
        yv_rgb = dkld65_to_rgb( torch.cat( [ white_dkl[0].expand_as(yv), white_dkl[1].expand_as(yv), yv ], dim=1 )+gray )

        frame = torch.cat( [ torch.cat( [ach_sust_rgb, ach_trans_rgb], dim=-1 ), 
                  torch.cat( [rg_rgb, yv_rgb], dim=-1 ) ], dim=-2 )
        for ff in range(frame.shape[2]): # for each frame
            frame_de = ((frame[0,:,ff,...] / self.max_V) ** (1/2.2) * 255).clip(0, 255)
            self.vw_channels.write_frame_rgb( frame_de.permute((1,2,0)).to(torch.uint8).cpu().numpy() )

    def dump_lpyr( self, lpyr, bands ):
        # Dump the content of the Laplacian pyramid for a given channel
        # Order: test-sustained-Y, ref-sustained-Y, test-rg, ref-rg, test-yv, ref-yv, test-transient-Y, ref-transient-Y
        # Images do not have the two last channels
        # R = torch.zeros((1, 8, cur_block_N_frames, height, width), device=self.device)

        # We dump only test at this moment

        if not self.do_dump_lpyr:
            return

        b0 = lpyr.get_band(bands, 0)
        b0_sh = b0.shape
        width = ceil8((b0_sh[-1] + lpyr.get_band(bands, 1).shape[-1] + 1)*2)
        height = ceil8((b0_sh[-2]+1)*2)
        frames = b0_sh[1]
        lpv = torch.zeros( [3, frames, height, width], device=b0.device)

        white_dkl = torch.as_tensor( [1, 0.003775328226986, 0.010327227989383], device=b0.device )
        gray = white_dkl.view([3,1,1,1])
        gray[0] /= 2

        B = lpyr.get_band_count()
        CHs = [0, 6, 2, 4]
        for col in range(4):
            ch = CHs[col]
            pos = [int(col/2)*int(height/2), (col%2)*int(width/2)]
            for bb in range(B):
                band = lpyr.get_band(bands,bb)[ch:(ch+1),...]
                if ch in (0,1,6,7): # achromatic
                    band_col = dkld65_to_rgb( torch.cat( [ band+white_dkl[0]/2, white_dkl[1].expand_as(band), white_dkl[2].expand_as(band) ], dim=0 ) )
                elif ch in (2,3): # RG
                    band_col = dkld65_to_rgb( torch.cat( [ white_dkl[0].expand_as(band)/2, band+white_dkl[1], white_dkl[2].expand_as(band) ], dim=0 ) )
                elif ch in (4,5): # YV
                    band_col = dkld65_to_rgb( torch.cat( [ white_dkl[0].expand_as(band)/2, white_dkl[1].expand_as(band), band + white_dkl[2] ], dim=0 ) )
                lpv[:, :, pos[0]:(pos[0]+band.shape[-2]), pos[1]:(pos[1]+band.shape[-1])] = band_col
                if (bb % 2)==0:
                    pos[1] += band.shape[-1]+1
                else:
                    pos[0] += band.shape[-2]+1

        for ff in range(frames): # for each frame
            frame_de = ((lpv[:,ff,...]) ** (1/2.2) * 255).clip(0, 255)
            self.vw_lpyr.write_frame_rgb( frame_de.permute((1,2,0)).to(torch.uint8).cpu().numpy() )

    def set_diff_band( self, width, height, pix_per_deg, bb, band):
        if not self.do_dump_diff:
            return

        if not self.diff_pyr:
            self.diff_pyr = lpyr_dec_2(width, height, pix_per_deg, band.device)
        self.diff_pyr.set_lband(bb, band)


    def dump_diff( self ):
        # Dump the per-band differences between the test and reference images
        # Order: test-sustained-Y, ref-sustained-Y, test-rg, ref-rg, test-yv, ref-yv, test-transient-Y, ref-transient-Y
        # Images do not have the two last channels
        # R = torch.zeros((1, 8, cur_block_N_frames, height, width), device=self.device)

        # We dump only test at this moment

        if not self.do_dump_diff:
            return

        b0 = self.diff_pyr.get_lband(0)
        b0_sh = b0.shape
        width = ceil8((b0_sh[-1] + self.diff_pyr.get_lband(1).shape[-1] + 1)*2)
        height = ceil8((b0_sh[-2]+1)*2)
        frames = b0_sh[1]
        lpv = torch.zeros( [3, frames, height, width], device=b0.device)

        white_dkl = torch.as_tensor( [1, 0.003775328226986, 0.010327227989383], device=b0.device )

        B = self.diff_pyr.get_band_count()
        CHs = [0, 3, 1, 2]
        for col in range(4):
            ch = CHs[col]
            pos = [int(col/2)*int(height/2), (col%2)*int(width/2)]
            for bb in range(B):
                band = self.diff_pyr.get_lband(bb)[ch:(ch+1),...]
                band_col = (band/10).expand( (3,-1,-1,-1) )
                lpv[:, :, pos[0]:(pos[0]+band.shape[-2]), pos[1]:(pos[1]+band.shape[-1])] = band_col
                if (bb % 2)==0:
                    pos[1] += band.shape[-1]+1
                else:
                    pos[0] += band.shape[-2]+1

        for ff in range(frames): # for each frame
            frame_de = ((lpv[:,ff,...]) ** (1/2.2) * 255).clip(0, 255)
            self.vw_diff.write_frame_rgb( frame_de.permute((1,2,0)).to(torch.uint8).cpu().numpy() )

    def close(self):
        if self.vw_channels:
            self.vw_channels.close()
        if self.vw_lpyr:
            self.vw_lpyr.close()
        if self.vw_diff:
            self.vw_diff.close()




import torch
import numpy as np 
import os
import sys
import math

def bucketize(tensor, bucket_boundaries):
    if tensor.device.type != 'mps':
        return torch.bucketize(tensor, bucket_boundaries)

    # MPS does not support bucketize yet
    result = torch.zeros_like(tensor, dtype=torch.int32)
    for boundary in bucket_boundaries:
        result += (tensor > boundary).int()
    assert (result.cpu() == torch.bucketize(tensor.cpu(), bucket_boundaries.cpu())).flatten().all()
    return result

# x_q : query tensor 
# x   : boundaries tensor
# inspired from: https://github.com/sbarratt/torch_interpolations/blob/master/torch_interpolations/multilinear.py#L39
def get_interpolants_v1(x_q, x):
    imax = bucketize(x_q, x)
    imax[imax >= x.shape[0]] = x.shape[0] - 1
    imin = (imax - 1).clamp(0, x.shape[0] - 1)

    ifrc = (x_q - x[imin]) / (x[imax] - x[imin] + 0.000001)
    ifrc[imax == imin] = 0.
    ifrc[ifrc < 0.0] = 0.

    return imin, imax, ifrc

def get_interpolants_v0(x_q, x, device):
    imin = torch.zeros(x_q.shape, dtype=torch.long).to(device)
    ifrc = torch.zeros(x_q.shape, dtype=torch.float32).to(device)
    N = x.shape[0]
    for i in range(N):
        if i==0:
            imin  = torch.where(x_q  <= x[i], torch.tensor(i, dtype=torch.long).to(device),  imin)
            ifrc  = torch.where(x_q  <= x[i], torch.tensor(0.).to(device), ifrc)

        if i==(N-1):
            imin  = torch.where(x[i] <= x_q,  torch.tensor(i, dtype=torch.long).to(device),  imin)
            ifrc  = torch.where(x[i] <= x_q,  torch.tensor(0.).to(device), ifrc)
        else:
            t = (x_q - x[i])/(x[i+1] - x[i])
            imin  = torch.where((x[i] <= x_q) & (x_q < x[i+1]), torch.tensor(i,dtype=torch.long).to(device), imin)
            ifrc  = torch.where((x[i] <= x_q) & (x_q < x[i+1]), t, ifrc)

    imax = torch.min(imin+1, torch.tensor(N-1, dtype=torch.long).to(device))

    return imin, imax, ifrc

# works only with uniformly sampled LUTs
def get_interpolants_quick(x_q, x):
    ind = ((x_q-x[0])/(x[-1]-x[0])*(x.numel()-1)).clamp(0,x.shape[0] - 1)
    ifrc = torch.frac(ind)
    imin = ind.to(dtype=torch.int32)
    imax = (imin+1).clamp(max=x.shape[0] - 1)    
    return imin, imax, ifrc


def interp3(x, y, z, v, x_q, y_q, z_q):
    shp = x_q.shape
    x_q = x_q.flatten()
    y_q = y_q.flatten()
    z_q = z_q.flatten()

    imin, imax, ifrc = get_interpolants_v1(x_q, x)
    jmin, jmax, jfrc = get_interpolants_v1(y_q, y)
    kmin, kmax, kfrc = get_interpolants_v1(z_q, z)

    filtered = (
        ((v[jmin,imin,kmin] * (1.0-ifrc) + v[jmin,imax,kmin] * (ifrc)) * (1.0-jfrc) + 
         (v[jmax,imin,kmin] * (1.0-ifrc) + v[jmax,imax,kmin] * (ifrc)) *     (jfrc)) * (1.0 - kfrc) + 
        ((v[jmin,imin,kmax] * (1.0-ifrc) + v[jmin,imax,kmax] * (ifrc)) *     (1.0-jfrc) + 
         (v[jmax,imin,kmax] * (1.0-ifrc) + v[jmax,imax,kmax] * (ifrc)) *     (jfrc)) * (kfrc))

    return filtered.reshape(shp)

def interp1(x, v, x_q):
    shp = x_q.shape
    x_q = x_q.flatten()

    imin, imax, ifrc = get_interpolants_v1(x_q, x)

    filtered = v[imin] * (1.0-ifrc) + v[imax] * (ifrc) 

    return filtered.reshape(shp)

# A quick interpolation for uniformly spaces samples
def interp1q(x, v, x_q):
    shp = x_q.shape
    x_q = x_q.flatten()

    imin, imax, ifrc = get_interpolants_quick(x_q, x)

    filtered = v[imin] * (1.0-ifrc) + v[imax] * (ifrc) 

    return filtered.reshape(shp)



# Performs 1-d interpolation on the 2nd dimension of tensor 'v' (hard coded as I do not know how to do it otherwise)
# This is equivalent to performing such interpolation multiple times, for each slice of other dimensions.
# x - tensor with the LUT x values, only one dimension can have size != 1
# v - tensor with the LUT v values (to be interpolated)
# x_q - tensor with the query x values, only one dimension can have size != 1
def interp1dim2(x, v, x_q):
    assert x.dim() == 1, "'x' must be a 1D vector"
    assert x_q.dim() == 1, "'x_q' must be a 1D vector"
    assert x.shape[0] == v.shape[1], "'x' must have the same number of elements as the second dimensiom of v"

    imin, imax, ifrc = get_interpolants_v1(x_q, x)
    sh = [1] * v.dim()
    sh[1] = ifrc.shape[0]
    ifrc = ifrc.view(sh)

    filtered = v[:,imin,...] * (1.0-ifrc) + v[:,imax,...] * (ifrc) 

    return filtered

def test_interp3(device):
    x_q = torch.tensor([0.5, 1.9, 2.1], dtype=torch.float32).to(device)
    y_q = torch.tensor([2.0, 2.0, 2.0], dtype=torch.float32).to(device)
    z_q = torch.tensor([1.5, 2.0, 2.0], dtype=torch.float32).to(device)

    x = torch.tensor([1.0, 2.0, 3.0], dtype=torch.float32).to(device)
    y = torch.tensor([1.0, 2.0, 3.0], dtype=torch.float32).to(device)
    z = torch.tensor([1.0, 2.0, 3.0], dtype=torch.float32).to(device)
    v = torch.tensor([
        [
        [10.0, 20.0, 30.0],
        [15.0, 30.0, 45.0],
        [20.0, 40.0, 60.0]],
        [
        [100.0, 200.0, 300.0],
        [150.0, 300.0, 450.0],
        [200.0, 400.0, 600.0]],
        [
        [1000.0, 2000.0, 3000.0],
        [1500.0, 3000.0, 4500.0],
        [2000.0, 4000.0, 6000.0]],
        ], dtype=torch.float32).to(device)

    print(x_q)
    print(x)
    print(v)
    print(interp3(x, y, z, v, x_q, y_q, z_q))


def batch_interp1d(x, xp, fp):
    """
    Perform batch-wise linear interpolation.
    """
    # Ensure xp is increasing
    assert torch.all(xp[1:] >= xp[:-1]), "xp must be in increasing order"

    # Make tensors contiguous to avoid warnings and optimize performance
    x = x.contiguous()
    xp = xp.contiguous()
    fp = fp.contiguous()

    # Find indices of the closest points
    indices = torch.searchsorted(xp, x) - 1
    indices = torch.clamp(indices, 0, len(xp) - 2)

    # Gather the relevant points
    x0 = xp[indices]
    x1 = xp[indices + 1]
    y0 = fp[torch.arange(fp.shape[0]), indices]
    y1 = fp[torch.arange(fp.shape[0]), indices + 1]

    # Compute the slope
    slope = (y1 - y0) / (x1 - x0)

    # Compute the interpolated values
    return y0 + slope * (x - x0)


if __name__ == '__main__':
    test_interp3(torch.device('cpu'))



# Classes for reading images or videos from files so that they can be passed to ColorVideoVDP frame-by-frame

from asyncio.log import logger
from functools import cache
import os
# from turtle import color
import imageio.v2 as io
import numpy as np
from torch.functional import Tensor
import torch
import ffmpeg
import re
import math
import torch.nn.functional as Func

import scipy.io as sio

import logging
from video_source import *
from video_source_yuv import video_reader_yuv

try:
    # This may fail if OpenEXR is not installed. To install,
    # ubuntu: sudo apt install libopenexr-dev
    # mac: brew install openexr
    import pyexr
    pyexr_imported = True
except ImportError as e:
    # Imageio's imread is unreliable for OpenEXR images
    # See https://github.com/imageio/imageio/issues/517
    pyexr_imported = False

# Load an image (SDR or HDR) into Numpy array
def load_image_as_array(imgfile):
    if not os.path.isfile(imgfile):
        msg = f"File '{imgfile}' not found"
        logger.error( msg )
        raise FileNotFoundError( msg )

    ext = os.path.splitext(imgfile)[1].lower()
    if ext == '.exr':
        if not pyexr_imported:
            logging.error( "pyexr is needed to read OpenEXR files. Please follow the instriction in README.md to install it." )
            raise RuntimeError( "pyexr missing" )
        precisions = pyexr.open(imgfile).precisions
        assert precisions.count(precisions[0]) == len(precisions), 'All channels must have same precision'
        img = pyexr.read(imgfile, precision=precisions[0])
    else:
        # 16-bit PNG not supported by default
        lib = 'PNG-FI' if ext == '.png' else None
        try:
            img = io.imread(imgfile, format=lib)
        except RuntimeError:
            logging.warning('PNG-FI not found, downloading using imageio\'s script')
            import imageio
            imageio.plugins.freeimage.download()
            img = io.imread(imgfile, format=lib)

    if img.ndim==3 and img.shape[2]>3:
        logging.warning(f'Input image {imgfile} has more than 3 channels (alpha?). Ignoring the extra channels.')
        img = img[:,:,:3]

    # Expand the tensor to [H,W,1] if we have a one-channel image
    if img.ndim==2:
        img = img[:,:,np.newaxis]

    return img


class video_reader:

    def __init__(self, vidfile, frames=-1, resize_fn=None, resize_height=-1, resize_width=-1, verbose=False):
        try:
            if vidfile.lower().endswith('.y4m'):
                probe = ffmpeg.probe(vidfile, count_frames=None)
            else:
                probe = ffmpeg.probe(vidfile)
        except:
            raise RuntimeError("ffmpeg failed to open file \"" + vidfile + "\"")

        # select the first video stream
        video_stream = next((stream for stream in probe['streams'] if stream['codec_type'] == 'video'), None)

        self.width = int(video_stream['width']) 
        self.src_width = self.width
        self.height = int(video_stream['height'])
        self.src_height = self.height
        self.color_space = video_stream['color_space'] if ('color_space' in video_stream) else 'unknown'
        self.color_transfer = video_stream['color_transfer'] if ('color_transfer' in video_stream) else 'unknown'
        self.in_pix_fmt = video_stream['pix_fmt']

        avg_fps_num, avg_fps_denom = [float(x) for x in video_stream['r_frame_rate'].split("/")]
        self.avg_fps = avg_fps_num/avg_fps_denom

        if 'nb_read_frames' in video_stream:
            frames_in_vstream = int(video_stream['nb_read_frames'])
        elif 'nb_frames' in video_stream: 
            frames_in_vstream = int(video_stream['nb_frames'])
        else:
            # Metadata may not contain total number of frames - this is the case of some VP9 videos
            if 'tags' in video_stream and 'DURATION' in video_stream['tags']:
                duration_text = video_stream['tags']['DURATION']
                hrs, mins, secs = map(float, duration_text.split(':'))
                duration = (hrs * 60 + mins) * 60 + secs
                frames_in_vstream = int(np.floor(duration * self.avg_fps))
            else:
                frames_in_vstream = -1; # Unspecified number of frames

        if frames==-1:
            self.frames = frames_in_vstream
        else:    
            self.frames = frames if frames_in_vstream==-1 else min( frames_in_vstream, frames ) # Use at most as many frames as passed in "frames" argument

        self._setup_ffmpeg(vidfile, resize_fn, resize_height, resize_width, verbose)
        self.curr_frame = -1

    def _setup_ffmpeg(self, vidfile, resize_fn, resize_height, resize_width, verbose):
        if any(f'p{bit_depth}' in self.in_pix_fmt for bit_depth in [10, 12, 14, 16]): # >8 bit
            out_pix_fmt = 'rgb48le'
            self.bpp = 6 # bytes per pixel
            self.dtype = np.uint16
        else:
            out_pix_fmt='rgb24' # 8 bit
            self.bpp = 3 # bytes per pixel
            self.dtype = np.uint8

        stream = ffmpeg.input(vidfile)
        if (resize_fn is not None) and (resize_width!=self.width or resize_height!=self.height):
            resize_mode = resize_fn if resize_fn != 'nearest' else 'neighbor'
            stream = ffmpeg.filter(stream, 'scale', resize_width, resize_height, flags=resize_mode)
            self.width = resize_width
            self.height = resize_height

        self.frame_bytes = int(self.width * self.height * self.bpp)

        log_level = 'info' if verbose else 'quiet'
        stream = ffmpeg.output(stream, 'pipe:', format='rawvideo', pix_fmt=out_pix_fmt).global_args( '-loglevel', log_level )
        #.global_args('-hwaccel', 'cuda', '-hwaccel_output_format', 'cuda') - no effect on decoding speed
        self.process = ffmpeg.run_async(stream, pipe_stdout=True)

    def get_frame(self):
        in_bytes = self.process.stdout.read(self.frame_bytes )
        if not in_bytes or (self.frames!=-1 and self.curr_frame == self.frames):
            return None
        in_frame = np.frombuffer(in_bytes, self.dtype)
        self.curr_frame += 1
        return in_frame       

    def unpack(self, frame_np, device):
        if self.dtype == np.uint8:
            assert frame_np.dtype == np.uint8
            frame_t_hwc = torch.tensor(frame_np, dtype=torch.uint8)
            max_value = 2**8 - 1
            frame_fp32 = frame_t_hwc.to(device).to(torch.float32)
        elif self.dtype == np.uint16:
            max_value = 2**16 - 1
            frame_fp32 = self._npuint16_to_torchfp32(frame_np, device)

        RGB = frame_fp32.reshape(self.height, self.width, 3) / max_value
        return RGB

    # Torch does not natively support uint16. A workaround is to pack uint16 values into int16.
    # This will be efficiently transferred and unpacked on the GPU.
    # logging.info('Test has datatype uint16, packing into int16')
    def _npuint16_to_torchfp32(self, np_x_uint16, device):
        max_value = 2**16 - 1
        assert np_x_uint16.dtype == np.uint16
        np_x_int16 = torch.tensor(np_x_uint16.astype(np.int16), dtype=torch.int16)
        torch_x_int32 = np_x_int16.to(device).to(torch.int32)
        torch_x_uint16 = torch_x_int32 & max_value
        torch_x_fp32 = torch_x_uint16.to(torch.float32)
        return torch_x_fp32

    # Delete or close if program was interrupted
    def __del__(self):
        self.close()

    def close(self):
        if hasattr(self, "process") and not self.process is None:
            self.process.stdout.close()
            self.process.kill() # We may wait forever if we do not read all the frames
            self.process = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, tb):
        self.close()


'''
Decode frames to Yuv, perform upsampling and color conversion with pytorch (on the GPU)
'''
class video_reader_yuv_pytorch(video_reader):
    def __init__(self, vidfile, frames=-1, resize_fn=None, resize_height=-1, resize_width=-1, verbose=False):
        super().__init__(vidfile, frames, resize_fn, resize_height, resize_width, verbose)

        y_channel_pixels = int(self.width*self.height)
        self.y_pixels = y_channel_pixels
        self.y_shape = (self.height, self.width)

        if self.chroma_ss == "444":
            self.frame_bytes = y_channel_pixels*3
            self.uv_pixels = y_channel_pixels
            self.uv_shape = self.y_shape
        elif self.chroma_ss == "420":
            self.frame_bytes = y_channel_pixels*3//2
            self.uv_pixels = int(y_channel_pixels/4)
            self.uv_shape = (int(self.y_shape[0]/2), int(self.y_shape[1]/2))
        else:
            raise RuntimeError("Unrecognized chroma subsampling.")

        if self.bit_depth > 8:
            self.frame_bytes *= 2

    def _setup_ffmpeg(self, vidfile, resize_fn, resize_height, resize_width, verbose):

        # if not any(f'p{bit_depth}' in self.in_pix_fmt for bit_depth in [10, 12, 14, 16]): # 8 bit
        #     raise RuntimeError('GPU decoding not implemented for bit-depth 8')

        re_grp = re.search(r'p\d+', self.in_pix_fmt)
        self.bit_depth = 8 if re_grp is None else int(re_grp.group().strip('p'))


        self.chroma_ss = self.in_pix_fmt[3:6]
        if not self.chroma_ss in ['444', '420']: # TODO: implement and test 422
            raise RuntimeError(f"Unrecognized chroma subsampling {self.chroma_ss}")

        if self.bit_depth>8: 
            self.dtype = np.uint16
            out_pix_fmt = f'yuv{self.chroma_ss}p{self.bit_depth}le'
        else:
            self.dtype = np.uint8
            out_pix_fmt = f'yuv{self.chroma_ss}p'

        # Resize later on the GPU
        if resize_fn is not None:
            self.resize_fn = resize_fn
            self.resize_height = resize_height
            self.resize_width = resize_width

        stream = ffmpeg.input(vidfile)
        log_level = 'info' if verbose else 'quiet'
        stream = ffmpeg.output(stream, 'pipe:', format='rawvideo', pix_fmt=out_pix_fmt).global_args( '-loglevel', log_level )
        self.process = ffmpeg.run_async(stream, pipe_stdout=True)

    def unpack(self, x, device):
        Y = x[:self.y_pixels]
        u = x[self.y_pixels:self.y_pixels+self.uv_pixels]
        v = x[self.y_pixels+self.uv_pixels:]

        Yuv_float = self._fixed2float_upscale(Y, u, v, device)

        if self.color_space=='bt2020nc':
            # display-encoded (PQ) BT.2020 RGB image
            ycbcr2rgb = torch.tensor([[1, 0, 1.47460],
                                        [1, -0.16455, -0.57135],
                                        [1, 1.88140, 0]], device=device)
        else:
            # display-encoded (sRGB) BT.709 RGB image
            ycbcr2rgb = torch.tensor([[1, 0, 1.402],
                                    [1, -0.344136, -0.714136],
                                    [1, 1.772, 0]], device=device)

        RGB = Yuv_float @ ycbcr2rgb.transpose(1, 0)
        if (hasattr(self, 'resize_fn')) and (self.resize_fn is not None) \
            and (self.height != self.resize_height or self.width != self.resize_width):
            RGB = torch.nn.functional.interpolate(RGB.permute(2,0,1)[None],
                                                  size=(self.resize_height, self.resize_width),
                                                  mode=self.resize_fn)
            RGB = RGB.squeeze().permute(1,2,0)

        return RGB.clip(0, 1)

    def _np_to_torchfp32(self, X, device):
        if X.dtype == np.uint8:
            return torch.tensor(X, dtype=torch.uint8).to(device).to(torch.float32)
        elif X.dtype == np.uint16:
            return self._npuint16_to_torchfp32(X, device)


    def _fixed2float_upscale(self, Y, u, v, device):
        offset = 16/219
        weight = 1/(2**(self.bit_depth-8)*219)
        Yuv = torch.empty(self.height, self.width, 3, device=device)

        Y = self._np_to_torchfp32(Y, device)
        Yuv[..., 0] = torch.clip(weight*Y - offset, 0, 1).reshape(self.height, self.width)

        offset = 128/224
        weight = 1/(2**(self.bit_depth-8)*224)

        uv = np.stack((u, v))
        uv = self._np_to_torchfp32(uv, device)
        uv = torch.clip(weight*uv - offset, -0.5, 0.5).reshape(1, 2, self.uv_shape[0], self.uv_shape[1])

        if self.chroma_ss=="420":
            # TODO: Replace with a proper filter.
            uv_upscaled = torch.nn.functional.interpolate(uv, scale_factor=2, mode='bilinear')
        else:
            uv_upscaled = uv

        Yuv[...,1:] = uv_upscaled.squeeze().permute(1,2,0)

        return Yuv


# Floor function that should be robust to the floating point precision issues
def safe_floor(x):
    x_f = math.floor(x)
    return x_f if (x-x_f)<(1-1e-6) else x_f+1

    

'''
Use ffmpeg to read video frames, one by one.
The readers are initialized on the first frame access - this allows to pickle an object before it us used (required by Pytorch Lightning)
'''
class video_source_video_file(video_source_dm):

    def __init__( self, test_fname, reference_fname, display_photometry='sdr_4k_30', config_paths=[], fps=None, frames=-1, full_screen_resize=None, resize_resolution=None, ffmpeg_cc=False, verbose=False, ignore_framerate_mismatch=False ):

        self.fs_width = -1 if full_screen_resize is None else resize_resolution[0]
        self.fs_height = -1 if full_screen_resize is None else resize_resolution[1]

        if test_fname.endswith('.yuv') and reference_fname.endswith('.yuv'):
            self.reader = video_reader_yuv
        else:
            self.reader = video_reader if ffmpeg_cc else video_reader_yuv_pytorch
        self.reference_vidr = None
        self.reference_fname = reference_fname
        self.test_fname = test_fname
        self.in_frames = frames
        self.full_screen_resize = full_screen_resize
        self.resize_resolution = resize_resolution
        self.ffmpeg_cc = ffmpeg_cc
        self.verbose = verbose
        self.fps = fps       
        self.ignore_framerate_mismatch = ignore_framerate_mismatch 

        super().__init__(display_photometry=display_photometry, config_paths=config_paths)


        # Resolutions may be different here because upscaling may happen on the GPU
        # if self.test_vidr.height != self.reference_vidr.height or self.test_vidr.width != self.reference_vidr.width:
        #     raise RuntimeError( f'Test and reference video sequences must have the same resolutions. Found: test {self.test_vidr.width}x{self.test_vidr.height}, reference {self.reference_vidr.width}x{self.reference_vidr.height}' )

        # self.last_test_frame = None
        # self.last_reference_frame = None

    def get_frame_count(self):
        self.init_readers()
        return self.frames

    def init_readers(self):
        if self.reference_vidr is None:
            self.reference_vidr = self.reader(self.reference_fname, self.in_frames, resize_fn=self.full_screen_resize, resize_width=self.fs_width, resize_height=self.fs_height, verbose=self.verbose)
            self.test_vidr = self.reader(self.test_fname, self.in_frames, resize_fn=self.full_screen_resize, resize_width=self.fs_width, resize_height=self.fs_height, verbose=self.verbose)

            if self.test_vidr.frames == -1 and self.reference_vidr.frames == -1:
                logging.error( "Neither test nor reference video contains meta-data with the number of frames. You need to specify it with '--nframes' argument" )
                raise RuntimeError("Unknown number of frames")

            if not self.ignore_framerate_mismatch: # We cannot use the logic below if we have fps mismatch. video_source_temp_resample_file will handle that.
                if self.test_vidr.frames == -1:
                    self.frames = self.reference_vidr.frames
                elif self.reference_vidr.frames == -1:
                    self.frames = self.test_vidr.frames
                else:
                    self.frames = min(self.test_vidr.frames,self.reference_vidr.frames)
                    if self.test_vidr.frames != self.reference_vidr.frames:
                        logging.warning( f"Test and reference videos contain different number of frames ({self.test_vidr.frames} and {self.reference_vidr.frames}). Comparing {self.frames} frames.")

            # self.frames = self.test_vidr.frames if self.in_frames==-1 else self.in_frames

            for vr in [self.test_vidr, self.reference_vidr]:
                if vr == self.test_vidr:
                    logging.debug(f"Test video '{self.test_fname}':")
                else:
                    logging.debug(f"Reference video '{self.reference_fname}':")
                if self.full_screen_resize is None:
                    rs_str = ""
                else:
                    rs_str = f"->[{self.resize_resolution[0]}x{self.resize_resolution[1]}]"
                if not self.ignore_framerate_mismatch:  
                    self.fps = vr.avg_fps if self.fps is None else self.fps
                    logging.debug(f"  [{vr.src_width}x{vr.src_height}]{rs_str}, colorspace: {vr.color_space}, color transfer: {vr.color_transfer}, fps: {self.fps}, pixfmt: {vr.in_pix_fmt}, frames: {self.frames}" )

            # if color_space_name=='auto':
            #     if self.test_vidr.color_space=='bt2020nc':
            #         color_space_name="BT.2020"
            #     else:
            #         color_space_name="sRGB"

            if not self.ignore_framerate_mismatch and self.test_vidr.avg_fps != self.reference_vidr.avg_fps:
                logging.error(f"Test and reference videos have different frame rates: test is {self.test_vidr.avg_fps} fps, reference is {self.reference_vidr.avg_fps} fps." )
                raise RuntimeError( "Inconsistent frame rates" )



        if self.test_vidr.color_transfer=="smpte2084" and self.dm_photometry.EOTF!="PQ":
            logging.warning( f"Video color transfer function ({self.test_vidr.color_transfer}) inconsistent with EOTF of the display model ({self.dm_photometry.EOTF})" )


    # Return (height, width, frames) touple with the resolution and
    # the length of the video clip.
    def get_video_size(self):
        self.init_readers()
        if hasattr(self.test_vidr, 'resize_fn') and self.test_vidr.resize_fn is not None:
            return (self.test_vidr.resize_height, self.test_vidr.resize_width, self.frames )
        else:
            return (self.test_vidr.height, self.test_vidr.width, self.frames )

    # Return the frame rate of the video
    def get_frames_per_second(self) -> int:
        self.init_readers()
        return self.fps
    
    # Get a test (reference) video frames as a single-precision luminance map
    # scaled in absolute inits of cd/m^2. 'frame' is the frame index,
    # starting from 0. 
    def get_test_frame( self, frame, device, colorspace="Y" ) -> Tensor:
        self.init_readers()
        #print( f"{self.test_fname} - {self.fs_width}x{self.fs_height}" )
        # if not self.last_test_frame is None and frame == self.last_test_frame[0]:
        #     return self.last_test_frame[1]
        L = self._get_frame( self.test_vidr, frame, device, colorspace )
        # self.last_test_frame = (frame,L)
        return L

    def get_reference_frame( self, frame, device, colorspace="Y" ) -> Tensor:
        self.init_readers()
        # if not self.last_reference_frame is None and frame == self.last_reference_frame[0]:
        #     return self.last_reference_frame[1]
        L = self._get_frame( self.reference_vidr, frame, device, colorspace )
        # self.reference_test_frame = (frame,L)
        return L

    def _get_frame( self, vid_reader, frame, device, colorspace ):        
        self.init_readers()

        if frame != (vid_reader.curr_frame+1):
            raise RuntimeError( 'Video can be currently only read frame-by-frame. Random access not implemented.' )

        frame_np = vid_reader.get_frame()

        if frame_np is None:
            raise RuntimeError( 'Could not read frame {}'.format(frame) )

        return self._prepare_frame(frame_np, device, vid_reader.unpack, colorspace)

    def _prepare_frame( self, frame_np, device, unpack_fn, colorspace="Y" ):
        frame_t_hwc = unpack_fn(frame_np, device)
        frame_t = reshuffle_dims( frame_t_hwc, in_dims='HWC', out_dims="BCFHW" )

        I = self.apply_dm_and_color_transform(frame_t, colorspace)

        return I


'''
This video source will resample the frames over time and can handle test and reference videos that have different frame rates. 
It currently handles only constant fps video. 
'''
class video_source_temp_resample_file(video_source_video_file):

    max_fps = 166 # upsample to at most this FPS

    def __init__( self, test_fname, reference_fname, display_photometry='sdr_4k_30', config_paths=[], frames=-1, full_screen_resize=None, resize_resolution=None, ffmpeg_cc=False, verbose=False ):
        super().__init__(test_fname, reference_fname, display_photometry=display_photometry, config_paths=config_paths, frames=-1, full_screen_resize=full_screen_resize, 
                         resize_resolution=resize_resolution, ffmpeg_cc=ffmpeg_cc, verbose=verbose, ignore_framerate_mismatch=True)


        super().init_readers()
        test_fps = self.test_vidr.avg_fps
        ref_fps = self.reference_vidr.avg_fps

        # First check if we can find an integer resampling rate
        if test_fps % 1 == 0 and ref_fps % 1 == 0:
            gcd = math.gcd(int(test_fps),int(ref_fps))
            self.resample_fps = min( test_fps * ref_fps/gcd, __class__.max_fps )
        else:
            self.resample_fps = __class__.max_fps

        test_frames_resampled = int( self.test_vidr.frames*self.resample_fps/test_fps )
        ref_frames_resampled = int( self.reference_vidr.frames*self.resample_fps/ref_fps )
        if self.test_vidr.frames==-1:
            frames_resampled = ref_frames_resampled
        elif self.reference_vidr.frames==-1:
            frames_resampled = test_frames_resampled
        else:
            frames_resampled = min( test_frames_resampled, ref_frames_resampled )

        self.frames = frames_resampled if frames==-1 else frames

        logger.info( f"Test fps: {test_fps}; reference fps: {ref_fps}. Resampling videos to {self.resample_fps} frames per second. {self.frames} frames will be processed." )
        if test_frames_resampled != ref_frames_resampled:
            logger.warning( f"Test and reference videos contain different number of frames after resampling ({test_frames_resampled} and {ref_frames_resampled}). Comparing {self.frames} frames." )

        self.cache_ind = [-1, -1]
        self.cache_frame = [None, None]
    
    # Return the frame rate of the video
    def get_frames_per_second(self):
        return self.resample_fps

    def get_video_size(self):
        return super().get_video_size()


    def _get_frame( self, vid_reader, frame, device, colorspace ):        

        frame_ind = int(safe_floor(frame/self.resample_fps * vid_reader.avg_fps))

        ce = 0 if vid_reader == self.test_vidr else 1

        if self.cache_ind[ce] == frame_ind:  # if quering the same frame in the source video, return the cache entry
            return self.cache_frame[ce]
        else:
            self.cache_ind[ce] = frame_ind
            self.cache_frame[ce] = super()._get_frame( vid_reader, frame_ind, device=device, colorspace=colorspace )
            #self.cache_frame[ce] = self.cache_frame[ce][...,4:-4,4:-4]  # Crop 4 pixels from all the sided because of the dark frame in the test videos
            return self.cache_frame[ce]            


'''
Load video frame-by-frame from image files. It can also handle single images.
'''
class video_source_image_frames(video_source_dm):
        
    def __init__( self, test_fname, reference_fname, fps=0, frame_range=None, display_photometry='sdr_4k_30', config_paths=[], full_screen_resize=None, resize_resolution=None, verbose=False ):

        super().__init__(display_photometry=display_photometry, config_paths=config_paths)        

        if not fps:
            fps = 0
        self.fps = fps
        self.video_size = None
        (self.test_fname, test_has_frame_no) = self.convert_c2python_format_str(test_fname)
        (self.reference_fname, ref_has_frame_no) = self.convert_c2python_format_str(reference_fname)

        if full_screen_resize:
            logging.error("full-screen-resize not implemented for images.")
            raise RuntimeError( "Not implemented" )

        if test_has_frame_no != ref_has_frame_no:
            logger.error( "Both test and reference names must contain `%0Nd` string to be replaced with a frame number" )
            raise RuntimeError( "Incorrect file names" )

        if (fps > 0) != test_has_frame_no:
            logger.error( "A valid frames-per-second number (--fps) must be provided when input are video frames, or fps should be zero for images." )
            raise RuntimeError( "Incorrect fps" )

        if fps==0:
            self.N = 1
            self.ff_name = self.test_fname # Name of the first frame
        else:
            # Check how many frames we have
            if not frame_range:
                frame_range = range(0, 10000)

            last_frame = 0
            frame_count = 0
            for nn in frame_range:
                if os.path.isfile( self.test_fname.format(nn) ) and os.path.isfile( self.reference_fname.format(nn) ):
                    last_frame = nn
                    frame_count += 1
                else:
                    break

            if frame_count == 0:
                logger.error( f"No frames found for {test_fname} and {reference_fname}" )
                raise RuntimeError( "No frames" )

            logger.info( f"{frame_count} frames found" )
            self.N = frame_count
            self.frame_range = frame_range[0:frame_count]
            self.ff_name = self.test_fname.format(self.frame_range[0])
        

    def convert_c2python_format_str( self, str ):
        if not hasattr( self, 'format_re' ):
            self.format_re = re.compile( r"%(\d)*d" )

        m = self.format_re.search( str )        
        if m:
            has_frame_no = True
            (beg, end) = m.span()
            new_str = str[0:beg] + '{:' + str[beg+1:end] + '}' + str[end:]
        else:
            has_frame_no = False
            new_str = str
        return (new_str, has_frame_no)            

    def get_frames_per_second(self):
        return self.fps
            
    # Return a [height width frames] vector with the resolution and
    # the number of frames in the video clip. [height width 1] is
    # returned for an image.     
    def get_video_size(self):
        if self.video_size is None:
            # Need to load first image to get the dimensions
            self.img_cache = load_image_as_array(self.ff_name)
            self.video_size = (self.img_cache.shape[0], self.img_cache.shape[1], self.N)

        return self.video_size

    def get_test_frame( self, frame, device, colorspace="Y" ) -> Tensor:
        if frame==0 and not self.img_cache is None: # Use cache to avoid loading the same image twice
            I = self._get_frame( self.test_fname, frame, device, colorspace, self.img_cache )
            self.img_cache = None
            return I
        else:
            return self._get_frame( self.test_fname, frame, device, colorspace)

    def get_reference_frame( self, frame, device, colorspace="Y" ) -> Tensor:
        return self._get_frame( self.reference_fname, frame, device, colorspace)

    def _get_frame(self, file_name, frame, device, colorspace, cache_img=None):

        if not cache_img is None: 
            img = cache_img
        else:
            if self.fps>0: # video
                frame_num = self.frame_range[frame]
                file_name = file_name.format(frame_num)
            img = load_image_as_array(file_name)

        img_torch = numpy2torch_frame(img, 0, device)
        I = self.apply_dm_and_color_transform(img_torch, colorspace)    
        return I

            # if not full_screen_resize is None:
            #     logging.error("full-screen-resize not implemented for images.")
            #     raise RuntimeError( "Not implemented" )
            # self.vs = video_source_array( img_test, img_reference, 0, dim_order='HWC', display_photometry=display_photometry, config_paths=config_paths )

            # hdr_extensions = [".exr", ".hdr"]
            # if extension in hdr_extensions:
            #     if self.vs.dm_photometry.EOTF != "linear":
            #         logging.warning('Use a display model with linear color space (EOTF="linear") for HDR images. Make sure that the pixel values are absolute.')
            # else:
            #     if self.vs.dm_photometry.EOTF == "linear":
            #         logging.warning('A display model with linear colour space should not be used with display-encoded SDR images.')



'''
The same functionality as to fvvdp_video_source_video_file, but preloads all the frames and stores in the CPU memory - allows for random access.
'''
class video_source_video_file_preload(video_source_video_file):
    
    def _get_frame( self, vid_reader, frame, device, colorspace ):        

        if not hasattr( self, "frame_array_tst" ):

            # Preload on the first frame
            self.frame_array_tst = [None] * self.frames
            for ff in range(self.frames):
                frame_np = self.test_vidr.get_frame()
                self.frame_array_tst[ff] = frame_np
                if ff==0:
                    mb_used = self.frame_array_tst[0].size * self.frame_array_tst[0].itemsize * self.frames * 2 / 1e6
                    logging.debug( f"Allocating {mb_used}MB in the CPU memory to store videos ({self.frames} frames)." )


            self.frame_array_ref = [None] * self.frames
            for ff in range(self.frames):
                frame_np = self.reference_vidr.get_frame()
                self.frame_array_ref[ff] = frame_np


        if vid_reader is self.test_vidr:
            frame_np = self.frame_array_tst[frame]
        else:
            frame_np = self.frame_array_ref[frame]

        if frame_np is None:
            raise RuntimeError( 'Could not read frame {}'.format(frame) )

        return self._prepare_frame(frame_np, device, vid_reader.unpack, colorspace)


'''
Load Matlab's .mat files
'''
class video_source_matlab( video_source_array ):

    def get_content( self, mat_struct ):
        for v_name in mat_struct:
            var = mat_struct[v_name]
            # We need a heuristic here - image needs to have more than 10 pixels - otherwise it is confused with other variables
            if isinstance( var, np.ndarray ) and var.ndim > 1 and var.ndim <= 4 and var.size>10:
                return var.astype(np.single) if var.dtype == np.double else var

        raise RuntimeError( 'Cannot find image or video data in the .mat file' )

    def __init__( self, test_fname, reference_fname, fps=None, display_photometry='sdr_4k_30', config_paths=[] ):
        test_mat = sio.loadmat(test_fname)
        ref_mat = sio.loadmat(reference_fname)

        if fps is None:
            fps = 30 if not 'fps' in test_mat.keys() else float(test_mat['fps'])

        test_cnt = self.get_content(test_mat)
        ref_cnt = self.get_content(ref_mat)

        if test_cnt.ndim != ref_cnt.ndim: # or (test_cnt.shape != ref_cnt.shape).any():
            raise RuntimeError( 'Matlab matrices must have the same number of dimensions and size.' )

        chn_no=1
        frame_no=1
        if test_cnt.ndim==2:
            dim_order="HW"
        elif test_cnt.ndim==4:
            dim_order="HWCF"
            chn_no=test_cnt.shape[2]
            frame_no=test_cnt.shape[3]
        elif test_cnt.ndim==3 and test_cnt.shape[-1]==3:
            dim_order="HWC"
            chn_no=test_cnt.shape[2]
        else:
            dim_order="HWF"
            frame_no=test_cnt.shape[2]

        logger.debug( f"Loaded matlab matrices: width={ref_cnt.shape[1]} height={ref_cnt.shape[0]} color_channels={chn_no} frames={frame_no} fps={fps}" )

        super().__init__( test_cnt, ref_cnt, fps, dim_order=dim_order, display_photometry=display_photometry, config_paths=config_paths, )


'''
Recognize whether the file is an image of video and wraps an appropriate video_source for the given content.
'''
class video_source_file(video_source):

    # fps==None - auto-detect, fps==0 - image, video otherwise
    def __init__( self, test_fname, reference_fname, display_photometry='sdr_4k_30', config_paths=[], frames=-1, frame_range=None, fps=None, full_screen_resize=None, resize_resolution=None, preload=False, ffmpeg_cc=False, verbose=False ):
        # these extensions switch mode to images instead
        image_extensions = [".png", ".jpg", ".gif", ".bmp", ".jpeg", ".ppm", ".tiff", ".tif", ".dds", ".exr", ".hdr"]

        # assert os.path.isfile(test_fname), f'Test file does not exists: "{test_fname}"'
        # assert os.path.isfile(reference_fname), f'Reference file does not exists: "{reference_fname}"'

        extension = os.path.splitext(test_fname)[1].lower()

        if extension == '.mat':
            self.vs = video_source_matlab(test_fname, reference_fname, fps=fps, display_photometry=display_photometry, config_paths=config_paths)
        elif extension in image_extensions:
            assert os.path.splitext(reference_fname)[1].lower() in image_extensions, 'Test is an image, but reference is a video'
            self.vs = video_source_image_frames(test_fname, reference_fname, fps=fps, frame_range=frame_range, display_photometry=display_photometry, config_paths=config_paths, full_screen_resize=full_screen_resize, resize_resolution=resize_resolution, verbose=verbose)


            # # if color_space_name=='auto':
            # #     color_space_name='sRGB' # TODO: detect the right colour space
            # img_test = load_image_as_array(test_fname)
            # img_reference = load_image_as_array(reference_fname)
            # if not full_screen_resize is None:
            #     logging.error("full-screen-resize not implemented for images.")
            #     raise RuntimeError( "Not implemented" )
            # self.vs = video_source_array( img_test, img_reference, 0, dim_order='HWC', display_photometry=display_photometry, config_paths=config_paths )

            # hdr_extensions = [".exr", ".hdr"]
            # if extension in hdr_extensions:
            #     if self.vs.dm_photometry.EOTF != "linear":
            #         logging.warning('Use a display model with linear color space (EOTF="linear") for HDR images. Make sure that the pixel values are absolute.')
            # else:
            #     if self.vs.dm_photometry.EOTF == "linear":
            #         logging.warning('A display model with linear colour space should not be used with display-encoded SDR images.')

        else:
            assert os.path.splitext(reference_fname)[1].lower() not in image_extensions, 'Test is a video, but reference is an image'
            vs_class = video_source_video_file_preload if preload else video_source_video_file
            self.vs = vs_class( test_fname, reference_fname, 
                                display_photometry=display_photometry, 
                                config_paths=config_paths,
                                frames=frames,   
                                fps=fps,
                                full_screen_resize=full_screen_resize, 
                                resize_resolution=resize_resolution, 
                                ffmpeg_cc=ffmpeg_cc, 
                                verbose=verbose )

    # Return (height, width, frames) touple with the resolution and
    # the length of the video clip.
    def get_video_size(self):
        return self.vs.get_video_size()

    # Return the frame rate of the video
    def get_frames_per_second(self) -> int:
        return self.vs.get_frames_per_second()
    
    # Get a pair of test and reference video frames as a single-precision luminance map
    # scaled in absolute inits of cd/m^2. 'frame' is the frame index,
    # starting from 0. 
    def get_test_frame( self, frame, device, colorspace="Y" ) -> Tensor:
        return self.vs.get_test_frame( frame, device, colorspace )

    def get_reference_frame( self, frame, device, colorspace="Y" ) -> Tensor:
        return self.vs.get_reference_frame( frame, device, colorspace )
